UNIT TrapInt;

(**********************************************************************

 Intercept Interrupts, provide stack for the interrupt procedure.

 (c) Copyright 1996,99 by Klaus Hartnegg,  hartnegg @ gmx.de

 Description
   Comfortable hooking of interrupt procedures, provides big stack

 Background
   Borland's Turbo Pascal and Borland Pascal 7 allow to declare
   Interrupt procedures. However these procedures must work with the
   stack that happens to be available at the time the procedure is
   called. This stack may be as tiny as 16 bytes. But passing a string
   to another pascal procedures requires 256 bytes. Passing 3 real
   numbers requires 15 bytes.

 Use
   Declare a procedure with the same arguments as TrapProcedure below.
   Call TrapInterrupt with interrupt number and procedure name as arguments.
   The procedure may read and change the variable reg and may set abort:=true.
   If abort is set to true, then the original interrupt handler is not called,
   otherwise it is called after your procedure.

   If you want to stop interception, call ReleaseInt.
   If you don't do it, it will automatically be done at
   end of your program by an exit procedure.

 Caveats
   If you use this to intercept hardware interrupts, your procedure
   must do port[$20]:=$20 before it exits with abort=true

   To prevent interrupts from being disabled longer than necessary,
   your procedure should call EnableInterrupts as first command.

   If your procedures does anything that requires interrupts to be active,
   it must call EnableInterrupts before that.

 Limitation
   only one interrupt can be hooked at a time.

 Changes
   30-Jan-1999 cleanup, don't enable interrupts, added comments

**********************************************************************)


INTERFACE

uses
  dos;

type
  TrapProcedure = procedure (var reg:registers; var abort:boolean);

var
  LostIntCounter : word;

Procedure TrapInterrupt (_IntNr:byte; _IntProc:TrapProcedure);
Procedure ReleaseInt;

Procedure EnableInterrupts;  inline ($FB);


IMPLEMENTATION

{ stack for interrupt handler }
const IntrStackSize = 16384;
var   IntrStack : array [1..IntrStackSize] of byte;
const IntrSS : word = Seg(IntrStack[IntrStackSize]);
      IntrSP : word = Ofs(IntrStack[IntrStackSize]);

var
  IntNr   : byte;
  IntProc : TrapProcedure;
  OldInt  : pointer;
  OldExit : pointer;
  SSSave, SPSave,
  CallIP, CallCS  : word;
  active : byte;
  abort  : boolean;
  PReg   : registers;


Procedure IntHandler; FAR; Assembler;
asm
  cmp   byte ptr @firstcall, 0 { check if initialization call }
  je    @install

  push  ds                     { save registers ax and ds on stack }
  push  ax

  mov   ax, seg @data          { set data segment of turbo pascal }
  mov   ds,ax

  cmp   byte ptr active,0      { test activity flag }
  jne   @notreentrant          { skip hooked procedure if already active }

  inc   byte ptr active        { set activity flag        }

  pop   [PReg.&ax]             { get AX and DS from stack ... }
  pop   [PReg.&ds]             { ...and store them           }

  pop   [CallIP]               { get and store stack contents ... }
  pop   [CallCS]
  pop   [PReg.Flags]           { ... mainly to get access to flags }

  mov   [PReg.&bx], bx         { store register contents  }
  mov   [PReg.&cx], cx
  mov   [PReg.&dx], dx
  mov   [PReg.&bp], bp
  mov   [PReg.&si], si
  mov   [PReg.&di], di
  mov   [PReg.&es], es

{ cli }                        { not required, interrupts already disabled }
  mov   [SSSave], ss           { store old stack pointer    }
  mov   [SPSave], sp
  mov   ss,[IntrSS]            { set stack for turbo pascal }
  mov   sp,[IntrSP]

  mov   byte ptr abort, 0      { set abort to false }

  mov   di,offset preg         { call pascal procedure }
  push  ds
  push  di
  mov   di,offset abort
  push  ds
  push  di
  call  dword ptr IntProc

  cli                          { disable interrupts if IntProc enabled them }
  mov   ss,[SSSave]            { restore old stack pointer }
  mov   sp,[SPSave]

  mov   es,[PReg.&es]          { reload registers from reg variable }
  mov   bp,[PReg.&bp]
  mov   di,[PReg.&di]
  mov   si,[PReg.&si]
  mov   dx,[PReg.&dx]
  mov   cx,[PReg.&cx]
  mov   bx,[PReg.&bx]

  push  [PReg.Flags]           { restore old stack contents }
  push  [CallCS]
  push  [CallIP]

  dec   byte ptr active        { clear activity flag }

  cmp   byte ptr abort,0       { test abort flag     }

  mov   ax,[PReg.&ax]          { restore registers AX and DS }
  push  [PReg.&ds]
  pop   ds

  jne   @abort

  jmp   dword ptr cs:[@IntAddr]{ call old interrupt procedure }

@abort:
  iret                         { return from interrupt }


@notreentrant:
  inc   word ptr LostIntCounter{ note that one interrupt was skipped }
  pop   ax
  pop   ds
  jmp   dword ptr cs:[@IntAddr]{ then call old interrupt handler }


@install:   { write address of original handler into code segment }
  inc   byte ptr @firstcall

  mov   ax, word ptr OldInt
  mov   word ptr @intaddr, ax
  mov   ax, word ptr OldInt+2
  mov   word ptr @intaddr+2, ax

  retf

@firstcall: db 0

@intaddr:   dw 0
            dw 0

end;



Procedure TrapInterrupt (_IntNr:byte; _IntProc:TrapProcedure);
begin
  IntNr   := _IntNr;
  IntProc := _IntProc;

  LostIntCounter := 0;

  GetIntVec (IntNr, OldInt);
  IntHandler; { first call does initialization }

  OldExit := ExitProc;
  ExitProc := @ReleaseInt;
  SetIntVec (IntNr, @IntHandler);
end;


Procedure ReleaseInt;
begin
  ExitProc := OldExit;
  SetIntVec (IntNr, OldInt);
end;


END.
