{*************************************************************}
{               Korzh's classes and procedures                }
{                   Copyright  Korzh.com                     }
{                http://www.korzh.com/delphi/                 }
{                  mailto:delphi@korzh.com                    }
{-------------------------------------------------------------}
{           String handling functions and procedures          }
{*************************************************************}
unit KStrings;

interface
uses SysUtils, Classes, Windows;

{$B-}

type
  EKStringsError = class(Exception)
  end;

//set list of symbols which are separators for such functions as IsSeparator or GetToken
//each separator is one character,e.g. calling SetSeparators('.,;')
//means that the characters '.', ',' and ';' will be used as separators
procedure SetSeparatorList(AList : string);

function IsSeparator(ch : char) : boolean;

//returns True if character ch is a digit
function IsDigit(ch : char) : boolean;

//returns True if character ch is a hexadecimal digit
function IsHexDigit(ch : char) : boolean;

function IsOctDigit(ch : char) : boolean;

//returns True if string S is a number
function IsNumber(const S : string) : boolean;

//returns True if string S is a real number
function IsRealNum(const S : string) : boolean;

//returns True if string S is an integer number
function IsIntNum(const S : string) : boolean;

//returns True if character ch is a space character
function IsSpace(ch : char) : boolean;

//returns True if character ch is a alpha character
function IsAlpha(ch : char) : boolean;

//returns True if string S can be an identifier, so it does not contain any separators
function IsIdentifier(const S : string) : boolean;

//returns True if string S has one of the "positive" values e.g. "yes" or "True"
function IsPositive(const S : string) : boolean;

//returns True if there are at least one letter (non symbol character) in the string
function AlphaExists(const S : string) : boolean;

//returns True if there are at least one letter (non symbol character) in the wide string
function AlphaExistsW(const S : WideString) : boolean;

//skip all spaces starting from position P in string S
procedure SkipSpaces(const S : string; var P : integer);

//skip all symbols listed in Symbols starting from position P in string S
procedure SkipSymbols(const S : string; var P : integer; Symbols : string);

//returns next token from the SStream and cut the first part of the SStream accordingly
//One token can be a separator or a sequence of characters between spaces
function GetToken(var SStream : string) : string;

{$IFNDEF WIN32}
function Trim(const S : string) : string;
{$ENDIF}

//returns the string with Count spaces
function Space(Count : integer) : string;

//returns string where character Symbol is replicated Count times
function Replicate(Symbol : char; Count : integer) : string;

//returns string padded with Symb character to a total length of Wide
//character Symb is added on the right side of S
function Padr(const S : string; Wide : integer; Symb : char) : string;

//returns string padded with Symb character to a total length of Wide
//character Symb is added on the left side of S
function Padl(const S : string; Wide : integer; Symb : char) : string;

//retruns Num'th position of SubStr in S
function PosNum(const SubStr, S : string;Num : integer) : integer;

//case insensitive Pos
function CPos(const SubStr, S : string) : integer;

//returns the position of num etry SubStr string into S from p position in S
function XPos(SubStr, S : string; P, Num : integer) : integer;

//returns the position of Num from the end etry of SubStr string into S starting from p position in S
function BXPos(SubStr, S : string; P, Num : integer) : integer;

//case unsensitive variant of XPos
function CXPos(SubStr, S : string; P, Num : integer) : integer;

//replaces all SWhat entries by SOn in string SSource
function ReplaceStr(const SSource, SWhat, SOn : string) : string;

//replaces all SWhat entries by SOn in string SSource
//Can be case sensative or case insensative (CaseSens parameter)
function XReplaceStr(const SSource, SWhat, SOn : string; CaseSens : boolean) : string;

//erases all occurences of symbol CH in the string S
function EraseSymbol(const S : string; CH : char) : string;

//returns left part of string S before separator string SSep
function GetBeforeSep(const SSep, S : string) : string;

//returns right part of string S after separator string SSep
function GetAfterSep(const SSep, S : string) : string;

//returns True if S starts with SubStr
function StartsWith(const SubStr, S : string) : boolean;

//returns True if S starts with SubStr
//Can be case sensative or case insensative (CaseSens parameter)
function XStartsWith(SubStr, S : string; CaseSens : boolean) : boolean;

//returns True if S is ending by SubStr
function EndsBy(const SubStr, S : string) : boolean;
function EndingBy(const SubStr, S : string) : boolean;

//returns string where first letter is in upper case
function CapitalizeFirst(const S : string ) : string;

//returns file name with path (if exists) without extention
//(from the start of name/path to the first dot)
function OnlyFileName(const S : string) : string;

//changes the extention of the file name to the given one (Ext - must started with dot)
function KChangeFileExt(const FileName, Ext : string) : string;

//extracts the extention of the file name (takes right part of name from the first dot)
//result extension is started with dot
function KExtractFileExt(const FileName : string) : string;

//returns index of string S in the string list Strings, -1 - if string was not found
//comparing is case insensative
function IndexOfString(Strings : TStrings; const S : string) : integer;

procedure InsertToStrings(Strings : TStrings; Line, Pos : integer; const S : string);

procedure DeleteFromStrings(Strings : TStrings; Line1, Pos1, Line2, Pos2 : integer);

//add '\' to path end if it does not exist
function FixPath(const Path : string) : string;

//concatenate two paths. Inserts '\' if it is necessary
function ConcPath(const Path1, Path2 : string) : string;

//concatenate two paths. Inserts Delim if it is necessary
function XConcPath(const Path1, Path2 : string; Delim : string) : string;

//concatenates two names through dot
function ConcNames(const Name1, Name2 : string) : string;

//returns the count of inclusions of character ch in string S
function CharCount(S : string;ch : char) : integer;

//the same as StrToInt but returns 0 for empty string
function XStrToInt(S : string) : integer;

//"safe" variant of StrToInt. Returns digital part of string as a number
//returns 0 for empty string
function SafeStrToInt(S : string) : integer;

//the same as StrToInt but returns 0 for unproper string (not integer)
function ZStrToInt(S : string) : integer;

//returns integer value by string which represent some hexadecimal value
function HexToInt(Value : string) : DWORD;

//returns integer value by string which represent some octadecimal value
function OctToInt(Value : string) : integer;

//converts string to float value; always use dot as decimal separator and return DefValue if string S is empty
function XStrToFloat(const S : string; DefValue : Extended) : Extended;

//converts boolean to string ; False -> '0', True -> '1'
function BoolToStr(B : boolean) : string;

//converts string to boolean ; 0, 'False' or emty string - is False; all other is True
function XStrToBool(const S : string) : boolean;

//the same as previous but returns True for empty string;
function NStrToBool(const S : string) : boolean;

//convert string to float value; returns 0 if string S is not a float number
function ZStrToFloat(S : string) : Extended;

//the same as FloatToStr but always use dot as decimal separator
function XFloatToStr(Value : Extended) : string;

//returns quoted string with any quote
function QuotedString(const S : string; Quote : char) : string;

//returns quoted string converted to C-style (with \ as special character)
function QuotedCString(const S : string) : string;

//returns unquoted string
function UnquotedString(const S : string; Quote : char) : string;

function GetCExtraSymbol(const S : string; var P : integer) : char;

function XUnquotedCString(const S : string; var P : integer) : string;

//returns unquoted C-style string (with \ as special character)
function UnquotedCString(const S : string) : string;

//returns delimited text by StringList
function DelimitedText(Strings : TStrings; Delimiter, Quote : char) : string;

//compose string list by some string
//each row in the string separated by Delimiter and may be quoted by quote
procedure ComposeText(const S : string; Delimiter, Quote : char; Strings : TStrings);

//compose string list by some string
//each row in the string separated by comma and quoted by "
procedure ComposeCommaText(const S : string; Strings : TStrings);

//returns one section from INI-like string list;
//SectionName - name of section; FromIndex - index of row which the section searching will be started from
function ReadSLSectionValues(Strings : TStrings;const SectionName : string; Section : TStrings) : integer;

function XGetSLSectionIndex(FromIndex : integer; Strings : TStrings;const SectionName : string) : integer;

//the same as previous but start searching Section from FromIndex row of the list
function XReadSLSectionValues(FromIndex : integer; Strings : TStrings;const SectionName : string; Section : TStrings) : integer;

//returns the names of all sections in string list
procedure ReadSLSections(Strings, Sections : TStrings);

procedure GetNameAndValueSep(const Sep, S : string; var AName, AValue : string);
procedure GetNameAndValue(const S : string; var AName, AValue : string);

//returns True if the string S is included into SList list
//useful to avoid long if-then-else constructions for strings comparing
//e.g. if StrInList(SomeStr, ['a1', 'a2', 'a3', 'a4']) then ...
//instead of if (SomeStr = 'a1') or (SomeStr = 'a2') or ...
function StrInList(const S : string; SList : array of string) : boolean;

//null terminated strings handling routins
function StrPCat(Dest : PChar;Source : string) : PChar;

procedure ClearPChar(Dest :PChar;Size : Word);

//returns True if two strings are equal one to another except symbols indicated in IgnoreSymbols list
//(any multiply symbols are equal to one symbol)
function EqualToSymbols(const S1, S2, IgnoreSymbols : string) : boolean;

//returns True if two strings are equal one to another except spaces
//(any multiply spaces are equal to one symbol)
function EqualToSpaces(const S1, S2 : string) : boolean;

function FitToTemplate(S, Template : string; CaseSens : boolean) : boolean;

function ContainsAsWholeWord(const S : string; SubStr : string) : boolean;

implementation

var
 aSeps : string = '';

function IsSpace(ch : char) : boolean;
begin
  Result := (ch in [' ', #13, #10, #9]);
end;

function IsDigit(ch : char) : boolean;
begin
  Result := (ch >= '0') and (ch <= '9');
end;

function IsHexDigit(ch : char) : boolean;
begin
  Result := IsDigit(ch) or
    ((ch >='A') and (ch <= 'F')) or ((ch >='a') and (ch <= 'f'));
end;

function IsOctDigit(ch : char) : boolean;
begin
  Result := (ch >= '0') and (ch <= '7');
end;

function IsNumber(const S : string) : boolean;
var
  I : integer;
begin
  if S = '' then
  begin
    Result := False;
    exit;
  end;
  for I := 1 to Length(S) do
    if (S[I] <'0') or (S[I] > '9') then
    begin
      Result := False;
      exit
    end;
  Result := True;
end;

{----------------------------------}
function IsIntNum(const S : string) : boolean;
begin
  try
    StrToInt(S);
    Result := True;
  except
    on EConvertError do Result := False;
  end;
end;

function IsRealNum(const S : string) : boolean;
begin
  try
    StrToFloat(S);
    Result := True;
  except
    on Exception do Result := False;
  end;
end;

function IsSeparator(ch : char) : boolean;
begin
  Result := Pos(ch, aSeps) > 0;
end;

function IsAlpha(ch : char) : boolean;
begin
  Result := ((ch >= '0') and (ch <= '9')) or
    ((ch >= 'A') and (ch <= 'Z')) or
    ((ch >= '_') and (ch <= 'z'));
end;

function IsIdentifier(const S : string) : boolean;
var
  I : integer;
begin
  Result := False;
  for I := 1 to Length(S) do
    if not IsAlpha(S[I]) then Exit;
  Result := True;
end;

function IsPositive(const S : string) : boolean;
begin
  Result := StrInList(AnsiLowerCase(S), ['on', 'yes', 'True', '1']);
end;
                 
function AlphaExists(const S : string) : boolean;
var
  I : integer;
begin
  for I := 1 to Length(S) do
    if ((S[I] >= 'A') and (S[I] <= 'Z')) or
       ((S[I] >= 'a') and (S[I] <= 'z')) or
       (S[I] > #127) then
    begin
      Result := True;
      Exit
    end;
  Result := False;
end;

function AlphaExistsW(const S : WideString) : boolean;
var
  I : integer;
begin
  for I := 1 to Length(S) do
    if not (Char(S[I]) in
      ['!', '"', '#', '$', '%', '&','''','(',')','*','+',',','-','.','/','0',
       '1','2','3','4','5','6','7','8','9',':',';','<','=','>','?','@','[','\',
       ']','^','_','`','{','|','}','~'])
    then begin
      Result := True;
      Exit
    end;
  Result := False;
end;

procedure SetSeparatorList(AList : string);
begin
  aSeps := AList;
end;

procedure SkipSpaces(const S : string; var P : integer);
begin
  while (P <= Length(S)) and IsSpace(S[P]) do inc(P);
end;

procedure SkipSymbols(const S : string; var P : integer; Symbols : string);
begin
  while (P <= Length(S)) and (Pos(S[P], Symbols) > 0) do inc(P);
end;

function GetToken(var SStream : string) : string;
var
  nStart, nEnd : integer;
begin
  Result := '';
  if SStream = '' then exit;
  nStart := 1;
  SkipSpaces(SStream, nStart);
  if IsSeparator(SStream[nStart]) then
  begin
    nEnd := nStart;
    while (nEnd <= Length(SStream)) and IsSeparator(SStream[nEnd]) do
     inc(nEnd);
  end
  else begin
    nEnd := nStart;
    while nEnd <= Length(SStream) do
    begin
      if IsSpace(SStream[nEnd]) or IsSeparator(SStream[nEnd]) then break;
      inc(nEnd);
    end;
  end;
  Result := Copy(SStream, nStart, nEnd-nStart);
  SStream := Copy(SStream, nEnd, Length(SStream)-nEnd+1);
end;

{$IFNDEF WIN32}
function Trim(const S : string) : string;
var
  I : integer;
  pos1,pos2 : integer;
begin
  if S = '' then
  begin
    Result := '';
    exit;
  end;
  I := 1;
  while IsSpace(S[I]) do inc(I);
  pos1 := I;
  I := Length(S);
  while isSpace(S[I]) do dec(I);
  pos2 := I;
  Result := Copy(S, pos1, pos2-pos1+1);
end;
{$ENDIF}


{-------------------------------------------------------------------------}
function Space(Count : integer) : string;
var
  I : integer;
begin
  Result := '';
  for I := 1 to Count do Result := Result + ' ';
end;

function Replicate(Symbol : char; Count : integer) : string;
var
  I : integer;
begin
  Result := '';
  if Count < 0 then exit;
  for I := 1 to Count do
    Result := Result + Symbol;
end;

function Padr(const S : string; Wide : integer; Symb : char) : string;
begin
  Result := Copy(S, 1, Wide) + Replicate(Symb, Wide - Length(S));
end;

function Padl(const S : string; Wide : integer; Symb : char) : string;
begin
  Result := Replicate(Symb, Wide - Length(S)) + Copy(S, 1, Wide);
end;

//retrun Num'th position of SubStr in S
function PosNum(const SubStr, S : string;Num : integer) : integer;
var
  Len, N, I : integer;
begin
  Result := 0; N := 0;
  if (S = '') or (SubStr = '') then Exit;
  Len := Length(SubStr);
  for I := 1 to Length(S) do
  begin
    if (S[I] = Substr[1]) and (Copy(S,I,Len) = Substr) then
    begin
      inc(N);
      if N = Num then
      begin
        Result := I;
        exit
      end;
    end;
  end;
end;

function XPos(SubStr, S : string; P, Num : integer) : integer;
var
  I, J, N : integer;
  Founded : boolean;
begin
  N := 0; Result := 0;
  if (P > Length(S)) or (S = '') or (SubStr = '') then Exit;

  if P < 1 then P := 1;

  for I := P to Length(S) do
  begin
    Founded := False;
    if S[I] = SubStr[1] then
    begin
      Founded := True;
      for J := 2 to Length(SubStr) do
        if (I + J - 1 > Length(S)) or (SubStr[J] <> S[I + J - 1]) then
        begin
          Founded := False;
          break;
        end;
    end;
    if Founded then
    begin
      inc(N);
      if Num = N then
      begin
        Result := I;
        Exit;
      end;
    end;
  end;
end;

function BXPos(SubStr, S : string; P, Num : integer) : integer;
var
  I, J, n : integer;
  Founded : boolean;
  ModP : integer;
begin
  n := 0; Result := 0;
  if ((P < 1) and (P <> -1)) or (S = '') or (SubStr = '') then Exit;
  //p=-1 means 'start from the end of string'

  ModP := Length(S);
  if (P > 0) and (P < ModP) then ModP := P;

  for I := ModP downto 1 do
  begin
    Founded := True;
    for J := 1 to Length(SubStr) do
      if SubStr[J] <> S[I + J - 1] then
      begin
        Founded := False;
        break;
      end;

    if Founded then
    begin
      inc(n);
      if Num = n then
      begin
        Result := I;
        exit;
      end;
    end;
  end;
end;

function CXPos(SubStr,S : string; P, Num : integer) : integer;
begin
  Result := XPos(AnsiLowerCase(Substr), AnsiLowerCase(S), p, num);
end;

function CPos(const SubStr, S : string) : integer;
begin
  Result := Pos(AnsiLowerCase(SubStr), AnsiLowerCase(S));
end;

function ReplaceStr(const SSource, SWhat, SOn : string) : string;
var
  LPos, NPos : integer;
begin
  Result := '';
  LPos := 1;
  NPos := XPos(SWhat, SSource, 1, 1);
  while NPos <> 0 do
  begin
    Result := Result + Copy(SSource, LPos, NPos - LPos);
    Result := Result + SOn;
    LPos := NPos + Length(SWhat);
    NPos := XPos(SWhat, SSource, LPos, 1);
  end;
  Result := Result + Copy(SSource, LPos, Length(SSource));
end;

function XReplaceStr(const SSource, SWhat, SOn : string; CaseSens : boolean) : string;
var
  LPos, NPos : integer;
  S, WhatStr : string;
begin
  Result := '';
  if not CaseSens then
  begin
    S := AnsiLowerCase(SSource);
    WhatStr := AnsiLowerCase(SWhat);
  end
  else
    S := SSource;
  LPos := 1;
  NPos := XPos(WhatStr, S, 1, 1);
  while NPos <> 0 do
  begin
    Result := Result + Copy(SSource, LPos, NPos - LPos);
    Result := Result + SOn;
    LPos := NPos + Length(WhatStr);
    NPos := XPos(WhatStr, S, LPos, 1);
  end;
  Result := Result + Copy(SSource, LPos, Length(SSource));
end;

function EraseSymbol(const S : string; CH : char) : string;
var
  I : integer;
begin
  I := 1;
  Result := '';
  while I <= Length(S) do
  begin
    if S[I] <> CH then
      Result := Result + S[I];
    inc(I);
  end;
end;

function GetBeforeSep(const SSep, S : string) : string;
var
  SepPos : Integer;
begin
  SepPos := Pos(SSep, S);
  if SepPos > 0 then
    Result := Trim(Copy(S, 1, SepPos - 1))
  else
    Result := S;
end;

function GetAfterSep(const SSep, S : string) : string;
var
  SepPos : Integer;
begin
  SepPos := Pos(SSep, S);
  if SepPos > 0 then
    Result := Trim(Copy(S, SepPos + Length(SSep), Length(S)))
  else
    Result := S;
end;

function StartsWith(const SubStr, S : string) : boolean;
var
  I : integer;
begin
  Result := False;
  if Length(SubStr) > Length(S) then Exit;
  for I := 1 to Length(SubStr) do
    if S[I] <> SubStr[I] then Exit;
  Result := True;
end;

function XStartsWith(SubStr, S : string; CaseSens : boolean) : boolean;
var
  I : integer;
begin
  if not CaseSens then
  begin
    SubStr := AnsiLowerCase(SubStr);
    S := AnsiLowerCase(S);
  end;

  for I := 1 to Length(SubStr) do
    if S[I] <> SubStr[I] then
    begin
      Result := False;
      Exit;
    end;
  Result := True;
end;

function EndingBy(const SubStr, S : string) : boolean;
begin
  Result := EndsBy(SubStr, S);
end;

function EndsBy(const SubStr, S : string) : boolean;
begin
  Result := CompareText(SubStr, Copy(S, Length(S) - Length(SubStr) + 1, Length(S))) = 0;
end;

function CapitalizeFirst(const S : string) : string;
begin
 if S <> '' then
 begin
   Result := AnsiLowerCase(S);
   Result := AnsiUpperCase(Result[1]) + Copy(Result, 2, Length(Result) - 1);
 end
 else
   Result := '';
end;

function OnlyFileName(const S : string) : string;
var
  I, P : integer;
begin
  P := 0;
  for I := Length(S) downto 1 do
  begin
    if S[I] = '.' then
      P := I;
    if S[I] = '\' then break;
  end;    
  if P > 0 then
    Result := Copy(S, 1, P - 1)
  else
    Result := S;
end;

function KChangeFileExt(const FileName, Ext : string) : string;
begin
  Result := OnlyFileName(FileName) + Ext;
end;

function KExtractFileExt(const FileName : string) : string;
var
  p : integer;
begin
  p := Pos('.', FileName);
  if p > 0 then
    Result := Copy(FileName, p, Length(FileName))
  else
    Result := '';
end;

function CharCount(S : string; ch : char) : integer;
var
  I : integer;
begin
  Result := 0;
  for I := 1 to Length(S) do
    if S[I] = ch then inc(Result);
end;

function XStrToInt(S : string) : integer;
begin
  if Trim(S) = '' then
    Result := 0
  else
    Result := StrToInt(S);
end;

function SafeStrToInt(S : string) : integer;
var
  I : integer;
begin
  try
    S := Trim(S);
    Result := StrToInt(S);
  except
    on EConvertError do
    begin
      if S = '' then
        Result := 0
      else begin
        I := 1;
        while isDigit(S[I]) do inc(I);
        Delete(S, I, Length(S));
        if S = '' then
          Result := 0
        else
          Result := StrToInt(S);
      end;
    end;
  end;
end;

function ZStrToInt(S : string) : integer;
begin
  try
    S := Trim(S);
    if S <> '' then
      Result := StrToInt(S)
    else   
      Result := 0;
  except
    Result := 0;
  end;
end;

function HexDigitToInt(Ch : char) : byte;
var
  sb : byte;
begin
  sb := ord(ch);
  if (sb >= ord('A')) and (sb <= ord('F')) then
    Result := sb - ord('A') + 10
  else if (sb >= ord('a')) and (sb <= ord('f')) then
    Result := sb - ord('a') + 10
  else if (sb >= ord('0')) and (sb <= ord('9')) then
    Result := sb - ord('0')
  else
    raise EKStringsError.Create(ch + ' is not a hex digit');
end;

function HexToInt(Value : string) : DWORD;
var
  I : integer;
  base : DWORD;
begin
  Result := 0;
  Value := UpperCase(Value);
  base := 1;
  for I := Length(Value) downto 1 do
  begin
    Result := Result + HexDigitToInt(Value[I])*base;
    base := base shl 4;
  end;
end;

function OctDigitToInt(Ch : char) : integer;
var
  sb : byte;
begin
  sb := ord(ch);
  if (sb >= ord('0')) and (sb <= ord('7')) then
    Result := sb - ord('0')
  else
    raise EKStringsError.Create(ch + ' is not an oct digit');
end;

function OctToInt(Value : string) : integer;
var
  I : integer;
  base : integer;
begin
  Result := 0;
  base := 1;
  for I := Length(Value) downto 1 do
  begin
    Result := Result + OctDigitToInt(Value[I])*base;
    base := base*8
  end;
end;

function XStrToFloat(const S : string; DefValue : Extended) : Extended;
var
  OldDecSep : char;
begin
  OldDecSep := DecimalSeparator;
  DecimalSeparator := '.';
  try
    if S = '' then
      Result := DefValue
    else
      Result := StrToFloat(S);
  finally
    DecimalSeparator := OldDecSep;
  end;
end;

function ZStrToFloat(S : string) : Extended;
begin
  try
    S := Trim(S);
    Result := StrToFloat(S);
  except
    Result := 0;
  end;
end;

function XFloatToStr(Value : Extended) : string;
var
  OldDecSep : char;
begin
  OldDecSep := DecimalSeparator;
  DecimalSeparator := '.';
  try
    Result := FloatToStr(Value);
  finally
    DecimalSeparator := OldDecSep;
  end;
end;

function BoolToStr(B : boolean) : string;
begin
  Result := IntToStr(ord(B));
end;

function XStrToBool(const S : string) : boolean;
begin
  if (S = '') or (CompareStr(S, 'false') = 0) or (S = '0') then
    Result := False
  else
    Result := True;
end;

function NStrToBool(const S : string) : boolean;
begin
  if (CompareStr(S, 'false') = 0) or (S = '0') then
    Result := False
  else
    Result := True;
end;

//add '\' to path end if it does not exist
function FixPath(const Path : string) : string;
begin
  if (Path <> '') and (Path[Length(Path)] <> '\') then
    Result := Path + '\'
  else
    Result := Path;
end;

//concatenates two paths
function ConcPath(const Path1, Path2 : string) : string;
begin
  if (Path1 = '') or (Path2 = '') or
     (Path1[Length(Path1)] = '\') or
     (Path2[1] = '\') then
    Result := Path1 + Path2
  else
    Result := Path1 + '\' + Path2;
end;

//concatenates two paths with Delim separator
function XConcPath(const Path1, Path2 : string; Delim : string) : string;
begin
  if (Path1 = '') or (Path2 = '') or
     (Path1[Length(Path1)] = '\') or
     (Copy(Path1, Length(Path1)-Length(Delim)+1, Length(Delim)) = Delim) or
     (Copy(Path2, 1, Length(Delim)) = Delim) then
    Result := Path1 + Path2
  else
    Result := Path1 + Delim + Path2;
end;

//concatenates two names through dot
function ConcNames(const Name1, Name2 : string) : string;
begin
  if Name1 = '' then
    Result := Name2
  else if Name2 = '' then
    Result := Name1
  else
    Result := Name1 + '.' + Name2;
end;

function QuotedString(const S : string; Quote : char) : string;
var
  I : integer;
begin
{  if S = '' then
  begin
    Result := '';
    exit;
  end;
}

  Result := Quote;
  I := 1;
  while I <= Length(S) do
  begin
    if  S[I] < ' ' then
    begin
      Result := Result + Quote;
      while (I <= Length(S)) and (S[I] < ' ') do
      begin
        Result := Result + '#' + IntToStr(ord(S[I]));
        inc(I);
      end;
      if I > Length(S) then exit;
      Result := Result + Quote
    end
    else begin
      Result := Result + S[I];
      if S[I] = Quote then
        Result := Result + Quote;
      inc(I);
    end
  end;
  Result := Result + Quote;
end;

function QuotedCString(const S : string) : string;
var
  I : integer;
begin
  {
  if S = '' then
  begin
    Result := '';
    exit;
  end;}
  Result := '"';
  I := 1;
  while I <= Length(S) do
  begin
    case S[I] of
      #10  : Result := Result + '\n';
      #9   : Result := Result + '\t';
      #11  : Result := Result + '\v';
      #8   : Result := Result + '\b';
      #13  : Result := Result + '\r';
      #12  : Result := Result + '\f';
      #7   : Result := Result + '\a';
      '''' : Result := Result + '\''';
      '"'  : Result := Result + '\"';
      '\'  : Result := Result + '\\';
      else
        if S[I] < ' ' then
          Result := Result + '\x' + IntToHex(ord(S[I]), 2)
        else
          Result := Result + S[I];
    end;
    Inc(I);
  end;
  Result := Result + '"';
end;

function UnquotedString(const S : string; Quote : char) : string;
var
  I : integer;
begin
  if (S = '') or (S[1] <> Quote) then
  begin
    Result := S;
    Exit;
  end;
  Result := ''; I := 2;
  while I < Length(S) do
  begin
    if S[I] = Quote then
    begin
      if S[I + 1] = Quote then
      begin
        Result := Result + Quote;
        inc(I, 2);
        continue;
      end
      else
        break;
    end
    else begin
      Result := Result + S[I];
      inc(I);
    end;
  end;
end;

function GetCExtraSymbol(const S : string; var P : integer) : char;
var
  PP : integer;
begin
  if (P >= Length(S)) or (S[P] <> '\') then
  begin
    Result := S[P];
    inc(P);
    Exit;
  end;

  case S[P + 1] of
    'n' : Result := #10;
    't' : Result := #9;
    'v' : Result := #11;
    'b' : Result := #8;
    'r' : Result := #13;
    'f' : Result := #12;
    'a' : Result := #7;
    'x' : begin
            inc(P, 2); PP := P;
            while (P <= Length(S)) and isHexDigit(S[P]) do inc(P);
            Result := chr(HexToInt(Copy(S, PP, P - PP)));
            Exit;
          end;
    '0'..'9' :
          begin
            inc(P, 1); PP := P;
            while (P <= Length(S)) and isOctDigit(S[P]) do inc(P);
            Result := chr(OctToInt(Copy(S, PP, P - PP)));
            Exit;
          end;
    else
      Result := S[P + 1];
  end;
  inc(P, 2);
end;

function XUnquotedCString(const S : string; var P : integer) : string;
begin
  if (S = '') or (S[P] <> '"') then
  begin
    Result := S;
    Exit;
  end;
  Result := '';inc(P);
  while (P <= Length(S)) and (S[P] <> '"') do
  begin
    if S[P] = '\' then
    begin
      Result := Result + GetCExtraSymbol(S, P);
      Continue;
    end;
    Result := Result + S[P];
    inc(P);
  end;
  if P <= Length(S) then inc(P);
end;

function UnquotedCString(const S : string) : string;
var
  P : integer;
begin
  P := 1;
  Result := XUnquotedCString(S, P);
end;

function DelimitedText(Strings : TStrings; Delimiter, Quote : char) : string;
var
  I : integer;
begin
  Result := '';
  for I := 0 to Strings.Count - 1 do
  begin
    if I <> 0 then Result := Result + Delimiter;
    Result := Result + QuotedString(Strings[I], Quote);
  end;
end;

procedure ComposeText(const S : string; Delimiter, Quote : char; Strings : TStrings);
var
  I, prev : integer;
  b : boolean;
begin
  prev := 0;
  b := False;
  Strings.Clear;
  if Trim(S) <> '' then
  begin
    for I := 1 to Length(S) do
    begin
      if S[I] = Quote then b := not b;
      if not b and (S[I] = Delimiter) then
      begin
        Strings.Add(UnquotedString(Trim(Copy(S, prev + 1, I - prev - 1)), Quote));
        prev := I;
      end;
    end;
    Strings.Add(UnquotedString(Trim(Copy(S, prev + 1, Length(S))), Quote));
  end;
end;

procedure ComposeCommaText(const S : string; Strings : TStrings);
begin
  ComposeText(S, ',', '"', Strings);
end;

function IndexOfString(Strings : TStrings; const S : string) : integer;
var
  I : integer;
begin
  for I := 0 to Strings.Count - 1 do
    if CompareText(S, Strings[I]) = 0 then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;  
end;

procedure InsertToStrings(Strings : TStrings; Line, Pos : integer; const S : string);
var
  LineStr : string;
begin
  LineStr := Strings[Line];
  Insert(S, LineStr, Pos);
  Strings[Line] := LineStr;
end;

procedure DeleteFromStrings(Strings : TStrings; Line1, Pos1, Line2, Pos2 : integer);
var
  I : integer;
  RightSideStr : string;
begin
  if Line2 < Line1 then Exit;
  RightSideStr := Copy(Strings[Line2], Pos2, Length(Strings[Line2]));
  Strings[Line1] := Copy(Strings[Line1], 1, Pos1 - 1) + RightSideStr;
  for I := Line1 + 1 to Line2 - 1 do
    Strings.Delete(I);
end;

function ReadSLSectionValues(Strings : TStrings;const SectionName : string; Section : TStrings) : integer;
var
  I : integer;
begin
  Section.Clear;
  I := 0;
  while (I < Strings.Count) and (CompareText(Trim(Strings[I]), '[' + SectionName + ']') <> 0) do inc(I);
  if I < Strings.Count then
  begin
    inc(I);
    while (I < Strings.Count) and
          ((Trim(Strings[I]) = '') or (Trim(Strings[I])[1] <> '[')) do
    begin
      if Trim(Strings[I]) <> '' then
        Section.Add(Strings[I]);
      inc(I);
    end;
  end;
  Result := I - 1;
  if (Section.Count > 0) and (Trim(Section[Section.Count - 1]) = '') then
    Section.Delete(Section.Count - 1);
end;

function XGetSLSectionIndex(FromIndex : integer; Strings : TStrings;const SectionName : string) : integer;
begin
  Result := FromIndex;
  while Result < Strings.Count do
  begin
    if CompareText(Trim(Strings[Result]), '[' + SectionName + ']') = 0 then Break;
    inc(Result);
  end;
  if Result = Strings.Count then
    Result := -1;

  if (Result = -1) and (FromIndex <> 0) then
    Result := XGetSLSectionIndex(0, Strings, SectionName);
end;

function XReadSLSectionValues(FromIndex : integer; Strings : TStrings;const SectionName : string; Section : TStrings) : integer;
var
  I : integer;
  S : string;
begin
  Section.Clear;

  I := XGetSLSectionIndex(FromIndex, Strings, SectionName);
  if I <> -1 then
  begin
    inc(I);
    while I < Strings.Count do
    begin
      S := Trim(Strings[I]);
      if S <> '' then
        if S[1] = '[' then Break;

      if S <> '' then
        Section.Add(Strings[I]);
      inc(I);
    end;
    Result := I - 1;
  end
  else
    Result := FromIndex;

  if Section.Count > 0 then
    if Trim(Section[Section.Count - 1]) = '' then
      Section.Delete(Section.Count - 1);

end;

procedure ReadSLSections(Strings, Sections : TStrings);
var
  I : integer;
  S : string;
begin
  I := 0;
  while I < Strings.Count do
  begin
    S := Trim(Strings[I]);
    if (S <> '') and (S[1] = '[') and (S[Length(S)] = ']') then
      Sections.Add(Copy(Strings[I], 2, Length(S)-2));
    inc(I);
  end;
end;

procedure GetNameAndValueSep(const Sep, S : string; var AName, AValue : string);
var
  P : integer;
begin
  P := Pos(Sep, S);
  if P > 0 then
  begin
    AName := Copy(S, 1, P - 1);
    AValue := Copy(S, P + 1, Length(S));
  end
  else begin
    AName := '';
    AValue := '';
  end;
end;

procedure GetNameAndValue(const S : string; var AName, AValue : string);
begin
  GetNameAndValueSep('=', S, AName, AValue);
end;

function StrInList(const S : string; SList : array of string) : boolean;
var
  I : integer;
begin
  for I := 0 to High(SList) do
    if CompareText(SList[I], S) = 0 then
    begin
      Result := True;
      exit;
    end;
  Result := False;
end;

function StrPCat(Dest : PChar;Source : string) : PChar;
begin
 Source := Source + #0;
 Result := StrCat(Dest,@Source[1]);
end;

procedure ClearPChar(Dest :PChar; Size : Word);
var
  I: word;
begin
  for I:= 0 to Size - 1 do
  Dest[I] := #0;
end;

function EqualToSymbols(const S1, S2, IgnoreSymbols : string) : boolean;
var
  I, J : integer;

  function Ignore(ch : char) : boolean;
  begin
    Result := Pos(ch, IgnoreSymbols) > 0;
  end;

  procedure SkipIgnoreSymbols;
  begin
    while (I <= Length(S1)) and (Ignore(S1[I])) do inc(I);
    while (J <= Length(S2)) and (Ignore(S2[J])) do inc(J);
  end;
begin
  I := 1; J := 1;
  while (I <= Length(S1)) or (J <= Length(S2)) do
  begin
    SkipIgnoreSymbols;
    if (I > Length(S1)) or (J > Length(S2)) or (S1[I] <> S2[J]) then
    begin
      Result := False;
      Exit;
    end;
    if Ignore(S1[I]) or Ignore(S2[J]) then
      SkipIgnoreSymbols
    else begin
      inc(I);
      inc(J);
    end;
  end;
  Result := True;
end;

function EqualToSpaces(const S1, S2 : string) : boolean;
begin
  Result := EqualToSymbols(S1, S2, ' ');
end;

function FitToTemplate(S, Template : string; CaseSens : boolean) : boolean;

  function TryTemplate(const S, Template : string) : boolean;
  var
    I, J, PJ : integer;
    P, N : integer;
    SubStr, NewStr, NewTemplate : string;

    function FindNextWildChar(CP : integer) : integer;
    begin
      Result := CP;
      while Result <= Length(Template) do
      begin
        if Template[Result] in ['*', '?'] then Exit;
        inc(Result);
      end;
    end;
  begin
    if ((S = '') and (Template = '')) or (Template = '*') then
    begin
      Result := True;
      Exit;
    end;

    I := 1;PJ := 1;
    repeat
      J := FindNextWildChar(PJ);
      if Copy(Template, I, J - PJ) <> Copy(S, I, J - PJ) then
      begin
        Result := False;
        Exit;
      end
      else
        inc(I, J - PJ);

      if J > Length(Template) then
      begin
        Result :=  J > Length(S);
        Exit;
      end;

      if Template[J] = '*' then
      begin
        inc(J);
        Result := J > Length(Template);
        if Result then Exit;

        P := FindNextWildChar(J);
        SubStr := Copy(Template, J, P - J);
        N := 1;
        repeat
          P := XPos(SubStr, S, I, N);
          if P > 0 then
          begin
            NewStr :=  Copy(S, P + Length(SubStr), Length(S));
            NewTemplate := Copy(Template, J + Length(SubStr), Length(Template));
            if TryTemplate(NewStr, NewTemplate) then
            begin
              Result := True;
              Exit;
            end;
          end
          else begin
            Result := False;
            Exit;
          end;
          inc(N);
        until False;
      end
      else if Template[J] = '?' then
        I := J + 1;
      PJ := J;  
    until False;
  end;

begin
  if CaseSens then
    Result := TryTemplate(S, Template)
  else
    Result := TryTemplate(AnsiLowerCase(S), AnsiLowerCase(Template));
end;

function ContainsAsWholeWord(const S : string; SubStr : string) : boolean;
var
  P, FP : integer;
begin
  P := Pos(SubStr, S);
  if P > 0 then
  begin
    FP := P + Length(SubStr);
    Result := ((P = 1) or not IsAlpha(S[P-1])) and ((FP > Length(S)) or not IsAlpha(S[FP]));
  end
  else
    Result := False;
end;

end.
